// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_app_Component_hpp
#define anna_app_Component_hpp

#include <vector>
#include <anna/core/util/Component.hpp>
#include <anna/app/Application.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace app {

/**
   Clase de la que heredan los componentes de aplicacion.

   Solo deberia haber una instancia de cada uno de los componentes, pero no podemos
   declararlos como heredados de anna::Singleton porque debemos dar la posiblidad de
   que el programador re-escriba su comportamiento mediante herencia.

   Todos los componentes se arrancan y paran automaticamente desde la aplicacion.

   El siguiente ejemplo muestra como obtener la instancia de un componente asociado a nuestra aplicacion:

   \code
      Clase* objeto = anna::app::functions::component <Clase> (FILE_LOCATION);

      ..... uso del objeto ....
   \endcode

   Si el componente \em 'Clase' no hubiera sido registrado (instanciado) en nuestra aplicacion el metodo
   template anna::component lanzara una excepcion.
*/
class Component : public anna::Component {
public:
  /**
     Destructor.
  */
  virtual ~Component();

  /**
     Conecta explicitamente este componente con la aplicacion. Sera necesario invocar a este metodo
     cuando instanciemos un componentes despues de comenzar la ejecucion de nuestra aplicacion y
     cuando el nuevo componente este completamente listo para su inicializacion.
  */
  void attach() throw(RuntimeException);

  /**
     Devuelve una cadena con la informacion mas relevante de esta instancia.
     \return Una cadena con la informacion mas relevante de esta instancia.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un documento XML con la informacion mas relevante de esta instancia.
     \param parent Nodo XML del que colgar la informacion referente a esta instancia.
     \return Un documento XML con la informacion mas relevante de esta instancia.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw();

protected:
  /**
     Estados en los que puede estar un componente.
     \see Component.
  */
  struct State { enum _v { Stopped, Starting, Running }; };

  /**
     Contructor.
     @param className Nombre lgico asociado a esta clase.
  */
  explicit Component(const char* className);

  /**
     Devuelve el estado de este componente.
     \return el estado de este componente.
  */
  State::_v getState() const throw() { return a_state; }

  /**
     Indica que el nombre de un componente que debe ser initializa antes que este.
     \param componentName Nombre de componente requerido por esta instancia.
     \warning Solo tendra efecto antes de inicializar el componente.
  */
  void addPredecessor(const char* componentName) throw();

  /**
     metodo que debemos implementar si la clase heredada necesita algn tipo de inicializacin.
     Este metodo se invocara automaticamente desde anna::Application::start.
  */
  void initialize() throw(RuntimeException);

  /**
     metodo que debemos implementar en la clase heredada para implementar el proceso de parada de
     esta clase.

     Este metodo debe implementar un metodo de parada controlada. Se invocara automaticamente
     desde el nucleo
  */
  void stop() throw() { a_state = State::Stopped; do_stop(); }

  /**
     metodo que debemos implementar en la clase heredada para implementar el proceso de parada de
     esta clase. Se invocara automaticamente desde el nucleo

     Este metodo implementa un metodo de parada no controlada.
  */
  virtual void kill() throw() { stop(); }

private:
  typedef std::vector <std::string>::iterator iterator;

  State::_v a_state;
  std::vector <std::string> a_predecessors;

  Component(const Component& other);
  iterator begin() throw() { return a_predecessors.begin(); }
  iterator end() throw() { return a_predecessors.end(); }
  const std::string& data(iterator& ii) throw() { return *ii; }

  /**
     metodo que debemos implementar si la clase heredada necesita algn tipo de inicializacin.
     Este metodo se invocara automaticamente desde anna::Application::start.
  */
  virtual void do_initialize() throw(RuntimeException) = 0;

  /**
     Metodo que debemos implementar en la clase heredada para tratar el proceso de parada de
     esta clase.
     Este metodo debe implementar un metodo de parada controlada. Se invocara automaticamente
     desde el nucleo
  */
  virtual void do_stop() throw() = 0;

  /**
   * Metodo que debemos implementar en la clase heredada para tratar el proceso de clonado de
   * esta clase desde el punto de vista del proceso original.
   * \warning Si se detecta algun error terminara abortara la ejecucion del proceso hijo.
   */
  virtual void do_cloneParent() throw(RuntimeException) { ; }

  /**
   * Metodo que debemos implementar en la clase heredada para tratar el proceso de clonado de
   * esta clase desde el punto de vista del proceso hijo.
   * \warning Si se detecta algun error terminara abortara la ejecucion de este proceso.
   */
  virtual void do_cloneChild() throw(RuntimeException) { ; }

  friend void Application::startComponents() throw(RuntimeException);
  friend void Application::stopComponents() throw(RuntimeException);
  friend void Application::attach(Component*) throw(RuntimeException);
  friend Application& Application::clone() throw(RuntimeException);
};

}
}

#endif
