// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_Singleton_hpp
#define anna_core_Singleton_hpp

#include <typeinfo>

#include <anna/core/mt/Mutex.hpp>
#include <anna/core/mt/Guard.hpp>

namespace anna {

/**
 Singleton pattern. Limits the use to a unique class instance.

 Default constructor for T must be private in order to avoid
 external instantiation. Following, an example using class A:


 On A class header file, we will have:
 \code

 ...
 ...

 #include <anna/core/Singleton.hpp>

 class A : public anna::Singleton <A> {
 public:
    // getters
    int getData () const { return a_data; }

    // setters
    void setOtherData (const ASD& otherData) { a_otherData = otherData; }

    // operators
       ....

    // methods
       ...

 private:
    int a_data;
    ASD a_otherData;

    A ();

    friend class functions::Singleton <A>;
 };

\endcode

 To use A class:

 \code
#include <A.h>

...
...

 void anyClass::anyFunction () throw (RuntimeException)
 {
    A& a (A::instantiate ());

    cout << a.getData () << endl;
 }

 \endcode
*/
template <class T> class Singleton {
public:
  /**
     @return Class instance provided on template creation.
  */
  static T& instantiate() { return *alloc(true); }

  /**
     Release the class instance provided in this template creation.
     If no instance has been created, this will be ignored.
     This method only have to be used during the program termination.
  */
  static void release() { alloc(false); }

private:
  static T* alloc(const bool allocate) {
    static Mutex mutex;
    static T* result(NULL);

    if(allocate == true) {
      if(result == NULL) {
        Guard guard(mutex, "Singleton<T>::allocate");

        if(result == NULL)
          result = new T;
      }
    } else {
      if(result != NULL) {
        Guard guard(mutex, "Singleton<T>::deallocate");

        if(result != NULL) {
          delete result;
          result = NULL;
        }
      }
    }

    return result;
  }
};

} //namespace anna

#endif

