// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_Mutex_hpp
#define anna_core_mt_Mutex_hpp

#include <pthread.h>

#include <anna/core/mt/Safe.hpp>

namespace anna {

/**
   Clase para implementar secciones criticas no-reentrantes. El uso general sera el siguiente:

   \code
      mutex.lock ();

      try {
       <implementacion de la seccion critica>

        mutex.unlock ();
      }
      catch (Exception&) {
        mutex.unlock ();
        .... tratamiento de la excepcion ...
      }
   \endcode

   La funcionalidad de esta clase solo estara disponible en aplicaciones multithread.

   @see anna::Guard
   @see anna::Thread
   @see anna::Semaphore
*/
class Mutex : public Safe {
public:
  struct Mode { enum _v { Recursive, Strict }; };
  /**
    Constructor.
  */
  explicit Mutex(const Mode::_v mode = Mode::Recursive);

  /**
    Destructor.
  */
  virtual ~Mutex();

  virtual void lock() throw(RuntimeException);

  virtual void unlock() throw();

  /**
   * Devuelve \em true en caso de que haya conseguido bloquear el mutex, o
   * \em false en otro caso.
   * \return \em true en caso de que haya conseguido bloquear el mutex, o
   * \em false en otro caso.
   */
  bool trylock() throw(RuntimeException);

  /**
   * Operador de conversin.
   * \return El \em pthread_mutex_t asociado a esta instancia.
   */
  operator const pthread_mutex_t*() const throw() {
#ifdef _MT
    return &a_id;
#else
    return NULL;
#endif
  }

private:
#ifdef _MT
  pthread_mutex_t a_id;
#endif

  Mutex(const Mutex& other);
};

/**
 * Macro que incorpora la definicin de un Mutex e implementa los mtodos \em lock y \em unlock
 * en la clase desde la que se invoca.
 *
 * \warning Debe invocarse desde la parte privada de la clase para asegurar que las secciones crticas de la
 * misma slo se activaran mediante anna::Guard<T>, que es la forma ms segura y fcil.
 */
#define anna_declare_mutex(ClassName) \
   anna::Mutex a_autoMutex; \
   void lock () throw (anna::RuntimeException) { a_autoMutex.lock (); } \
   void unlock () throw () { a_autoMutex.unlock (); } \
   friend class anna::Guard <ClassName >;

#define anna_access_mutex a_autoMutex

} //namespace anna

#endif
