// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_tracing_Logger_hpp
#define anna_core_tracing_Logger_hpp

#include <syslog.h>

#ifdef _MT
#include <pthread.h>
#endif

#include <anna/core/mt/NRMutex.hpp>
#include <anna/core/mt/Guard.hpp>

namespace anna {

class DataBlock;

/**
   Facilidad para la realizacion de archivos de historico (logs) de nuestra aplicacion.

   @see Patron proxy  http://maniaco/anna/docs/html/DesingPatterns//pat4gfso.htm
*/
class Logger {
public:
  /**
     Nivel de las trazas de historico.

     Para mas informacion ver los niveles de emergencia listados en: man syslog.conf

     @see Logger
  */
  enum Level {
    Emergency = LOG_EMERG,
    Alert = LOG_ALERT, Critical = LOG_CRIT, Error = LOG_ERR, Warning = LOG_WARNING,
    Notice = LOG_NOTICE, Information = LOG_INFO, Debug = LOG_DEBUG,
    Local0 = LOG_LOCAL0, Local1 = LOG_LOCAL1, Local2 = LOG_LOCAL2, Local3 = LOG_LOCAL3,
    Local4 = LOG_LOCAL4, Local5 = LOG_LOCAL5, Local6 = LOG_LOCAL6, Local7 = LOG_LOCAL7
  };

  /**
     Clase virtual que debemos asociar al Logger para transferir las trazas al medio
     deseado.

     @see DefaultWriter
  */
  class Writer {
  public:
    /**
       Destructor.
    */
    virtual ~Writer();

    /**
       Metodo virtual que debe inicializar el funcionamiento del Writer particular.

       @param id Identificador logico. Usado en el  DefaultWriter para inicializar el sistema
       del log (metodo syslog).
    */
    virtual void initialize(const char* id) throw() = 0;

    /**
       Transfiere el mensaje de trazas al log del sistema. El metodo implementado debe ser
       MT-Safe.
       No es necesario realizar ningun control del nivel de trazas establecido ya que si la
       clase  Logger ha decidido llamar a este metodo es porque el nivel indicado esta
       activo.

       @param level Nivel de las trazas que vamos a sacar.
       @param text Especificacion de formato del mensaje. Similar al usado en el comando sprintf.
    */
    virtual void do_write(int level, const char* text, ...) throw() = 0;

  protected:
    /**
       Constructor.

       @param bufferSize Numero de bytes reservado para la linea de trazas.
    */
    Writer(const int bufferSize);

    /**
       Constructor.
    */
    Writer();

    /**
       @return La instancia del bloque de datos con memoria reservada para poder interpretar los
       parametros.
    */
    DataBlock& getDataBlock() throw() { return *a_dataBlock; }

  private:
    DataBlock* a_dataBlock;
  };

  /**
     Inicializa el sistema de historico de nuestra aplicacion. Solo debe invocarse una unica vez
     al comienzo de la aplicacion.

     Al no indicar ninguna clase encargada de transferir los datos de la aplicacion al
     sistema de historico se establece el  Writer por defecto.

     Este metodo no es MT-safe por lo que tenemos que estar seguros que no se puede invocar desde
     varios thread simultaneamente. Lo mas aconsejable es invocarlo desde el comienzo de la
     aplicacion.

     @param ident Identifica las trazas de nuestra aplicacion que apareceran  en el archivo de historico.
     Deberia ser un texto con entre 4 y 16 caracteres.

     @see DefaultWriter.
  */
  static void initialize(const char* ident) throw();

  /**
     Inicializa el sistema de historico de nuestra aplicacion. Solo debe invocarse una unica vez
     al comienzo de la aplicacion.

     Este metodo no es MT-safe por lo que tenemos que estar seguros que no se puede invocar desde
     varios thread simultaneamente. Lo mas aconsejable es invocarlo desde el comienzo de la
     aplicacion.

     @param ident Identifica las trazas de nuestra aplicacion que apareceran  en el archivo de historico.
     Deberia ser un texto con entre 4 y 16 caracteres.
     @param writer Establece el objeto encargado de transferir los datos de la aplicacion
     al sistema de historico. La instancia pasada como parametro debera estar disponible mientras
     no termine la ejecucion de nuestra aplicacion. Podemos indicar NULL para desactivar
     completamente cualquier sistema de trazas.

     @see DefaultWriter.
  */
  static void initialize(const char* ident, Writer* writer) throw();

  /**
     @return El nivel de trazado de nuestra aplicacion.
  */
  static Level getLevel() throw() { return st_level; }

  /**
     Establece el nivel de trazado de nuestra aplicacion. El nivel de trazado por defecto
     de nuestra aplicacion dependera del modo de compilacion, en modo depuracion el nivel
     de trazado por defecto sera  Debug, en otro caso sera  Warning.

     Solo apareceran en el historico las trazas que lleven un nivel menor que el establecido
     en este metodo.

     @param level Nivel de trazado que deseamos establecer.
  */
  static void setLevel(const Level level)
  throw(RuntimeException) {
    Guard guard(st_mutex, "Logger::setLevel");
    st_level = (level <= Error) ? Error : level;
  }

  /**
     Comprueba si el nivel de trazado recibido como parametro esta activo en nuestra aplicacion.

     @param level Nivel de trazado que deseamos comprobar.

     @return @em true Si el nivel de trazado de nuestra aplicacion es mayor que el recibido como
     parametro o @em false en otro caso.
  */
  static bool isActive(const Level level) throw() {
    return (st_writer != NULL && level <= Error) ? true : (st_enabled && level <= st_level && st_writer != NULL);
  }

  /**
     Desactiva el sistema de trazado.
  */
  static void disable() throw(RuntimeException);

  /**
     Activa el sistema de trazado y establece el nivel de trazado existente antes
     de la desactivacion (Ver  #disable).
  */
  static void enable() throw(RuntimeException);

  /**
   * Establece el valor del indicador que hace que se vuelque el PID del proceso en la linea de trazas.
   *
   * Puede ser util cuando el programa principal cree procesos hijos mediante a la invocacion al metodo \em fork.
   *
   * \param showPID Valor del indicador.
   */
  static void showPID(const bool show) throw();

  /**
     Traza el texto recibido en el historico con el nivel indicado.

     La traza solo sera registrada en el historico si el nivel de trazado recibido como
     parametro esta habilitado.

     @param level Nivel de la traza que deseamos registrar.
     @param text Texto de la traza.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void write(const Level level, const char* text, const char* fromFile, const int fromLine) throw();

  /**
     Traza el texto recibido en el historico con el nivel indicado.

     La traza solo sera registrada en el historico si el nivel de trazado recibido como
     parametro esta habilitado.

     @param level Nivel de la traza que deseamos registrar.
     @param text Texto de la traza.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void write(const Level level, const std::string& text, const char* fromFile, const int fromLine) throw() {
    write(level, text.c_str(), fromFile, fromLine);
  }

  /**
     Traza el texto recibido en el historico con el nivel indicado.

     La traza solo sera registrada en el historico si el nivel de trazado recibido como
     parametro esta habilitado.

     @param level Nivel de la traza que deseamos registrar.
     @param text Texto de la traza.
     @param value Contenido de una cadena.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void write(const Level level, const char* text, const char* value, const char* fromFile, const int fromLine) throw();

  /**
     Traza el texto recibido en el historico con el nivel indicado.

     La traza solo sera registrada en el historico si el nivel de trazado recibido como
     parametro esta habilitado.

     @param level Nivel de la traza que deseamos registrar.
     @param text Texto de la traza.
     @param value Contenido de una cadena.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void write(const Level level, const char* text, const std::string& value, const char* fromFile, const int fromLine)
  throw() {
    write(level, text, value.c_str(), fromFile, fromLine);
  }

  /**
     Traza el texto recibido en el historico con el nivel indicado.

     La traza solo sera registrada en el historico si el nivel de trazado recibido como
     parametro esta habilitado.

     @param level Nivel de la traza que deseamos registrar.
     @param text Texto de la traza.
     @param value Contenido de una cadena.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void write(const Level level, const std::string& text, const std::string& value, const char* fromFile, const int fromLine)
  throw() {
    write(level, text.c_str(), value.c_str(), fromFile, fromLine);
  }

  /**
     Traza el texto recibido en el historico con el nivel indicado.

     La traza solo sera registrada en el historico si el nivel de trazado recibido como
     parametro esta habilitado.

     @param level Nivel de la traza que deseamos registrar.
     @param text Texto de la traza.
     @param value Valor numerico.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void write(const Level level, const char* text, const int value, const char* fromFile, const int fromLine)
  throw();

  /**
     Traza el texto recibido en el historico con el nivel indicado.

     La traza solo sera registrada en el historico si el nivel de trazado recibido como
     parametro esta habilitado.

     @param level Nivel de la traza que deseamos registrar.
     @param text Texto de la traza.
     @param value Bloque de datos a transferir al log del sistema.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void write(const Level level, const char* text, const DataBlock& value, const char* fromFile, const int fromLine)
  throw();

  /**
     Si el nivel \em Debug esta activado traza el texto recibido en el historico.
     @param text Texto de la traza.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void debug(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Debug, text, fromFile, fromLine);
  }

  /**
     Si el nivel \em Information esta activado traza el texto recibido en el historico.
     @param text Texto de la traza.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void information(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Information, text, fromFile, fromLine);
  }

  /**
     Si el nivel \em Notice esta activado traza el texto recibido en el historico.
     @param text Texto de la traza.
     @param fromFile Nombre del archivo donde se genera la traza.
     @param fromLine Numero de linea del archivo donde se genera la traza.
  */
  static void notice(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Notice, text, fromFile, fromLine);
  }

  /**
      Si el nivel \em Warning esta activado traza el texto recibido en el historico.
      @param text Texto de la traza.
      @param fromFile Nombre del archivo donde se genera la traza.
      @param fromLine Numero de linea del archivo donde se genera la traza.
   */
  static void warning(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Warning, text, fromFile, fromLine);
  }

  /**
      Si el nivel \em Error esta activado traza el texto recibido en el historico.
      @param text Texto de la traza.
      @param fromFile Nombre del archivo donde se genera la traza.
      @param fromLine Numero de linea del archivo donde se genera la traza.
   */
  static void error(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Error, text, fromFile, fromLine);
  }

  /**
      Si el nivel \em Critical esta activado traza el texto recibido en el historico.
      @param text Texto de la traza.
      @param fromFile Nombre del archivo donde se genera la traza.
      @param fromLine Numero de linea del archivo donde se genera la traza.
   */
  static void critical(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Critical, text, fromFile, fromLine);
  }

  /**
      Si el nivel \em Alert esta activado traza el texto recibido en el historico.
      @param text Texto de la traza.
      @param fromFile Nombre del archivo donde se genera la traza.
      @param fromLine Numero de linea del archivo donde se genera la traza.
   */
  static void alert(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Alert, text, fromFile, fromLine);
  }

  /**
      Si el nivel \em Emergency esta activado traza el texto recibido en el historico.
      @param text Texto de la traza.
      @param fromFile Nombre del archivo donde se genera la traza.
      @param fromLine Numero de linea del archivo donde se genera la traza.
   */
  static void emergency(const std::string& text, const char* fromFile, const int fromLine)
  throw() {
    write(Logger::Emergency, text, fromFile, fromLine);
  }

  /**
     @return La cadena que identifica al nivel recibido como parametro.

  */
  static const char* asString(const Level level) throw();

  /**
     Traduce la cadena recibida al nivel correspondiente.

     @param level Cadena que deberia contener un nombre de nivel (emerg, alert, crit, err, warning, notice, info, debug).

     \warning Debe de ser alguno de los siguiente literales: emerg, alert, crit, err, warning, notice, info, debug
  */
  static Level asLevel(const char* level) throw(RuntimeException);

private:
  static NRMutex st_mutex;
  static Level st_level;
  static bool st_enabled;
  static Writer* st_writer;

#ifndef _MT
  static pid_t st_pid;
#endif

  Logger();

  friend class Logger::Writer;
};

} //namespace anna

#endif


