// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_comm_ServerSession_hpp
#define anna_diameter_comm_ServerSession_hpp


// STL
#include <string>

#include <anna/core/util/Millisecond.hpp>
#include <anna/core/RuntimeException.hpp>

#include <anna/diameter.comm/Session.hpp>
#include <anna/diameter/defines.hpp>
#include <anna/diameter.comm/Message.hpp>
#include <anna/diameter.comm/ServerSessionReceiver.hpp>
#include <anna/diameter.comm/ReceiverFactoryImpl.hpp>


namespace anna {
class DataBlock;
namespace timex {
class Engine;
}

namespace comm {
class ClientSocket;
}
}


namespace anna {

namespace diameter {

namespace comm {


class LocalServer;


/**
   Modela la conexion realizada contra un servidor diameter local.
*/
class ServerSession : public Session {
public:

  ServerSession();


  /* virtual */void initialize() throw();

  /**
   * Default max inactivity period for the diameter server-session health.
   */
  static const anna::Millisecond DefaultAllowedInactivityTime;

  /**
     Sets the maximum allowed inactivity time on server session

     @param allowedInactivityTime Inactivity time allowed
  */
  void setAllowedInactivityTime(const anna::Millisecond & allowedInactivityTime) throw();

  /**
     Diameter listening address (ip or hostname).
     \return Diameter listening address.
  */
  /* virtual */const std::string& getAddress() const throw();

  /**
     Diameter listen port.
     \return Diameter listen port.
  */
  /* virtual */int getPort() const throw();

  /**
     Server session key. Same as socket id
  */
  int getKey() const throw() { return getSocketId(); }

  /**
     Local server parent
     \return Local server parent
  */
  LocalServer *getParent() throw() { return a_parent; }


  /**
     Sets the diameter client socket and assign the receiver factory to it
     \param clientSocket Diameter client socket
  */
  void setClientSocket(anna::comm::ClientSocket *clientSocket) throw();

  /**
     Diameter client socket
     \return Diameter client socket
  */
  anna::comm::ClientSocket *getClientSocket() throw() { return a_clientSocket; }

//   /**
//      Sets deprecated state to this server session
//   */
//   void setDeprecated(bool deprecated = true) throw() { a_deprecated = deprecated; }

//   /**
//    * Sets CEA and DWA diameter messages to be used over created server-sessions
//    *
//    * @param cea Capabilities-Exchange-Answer message (encoded) for the server-sessions bind.
//    * @param dwa Device-Watchdog-Answer message (encoded) for the server-sessions keep-alive.
//    */
//   void setCEAandDWA(const anna::DataBlock & cea, const anna::DataBlock & dwa) throw(anna::RuntimeException);

  /* virtual */const Response* send(const Message* message) throw(anna::RuntimeException);
  /* virtual */bool unbind(bool forceDisconnect /* se usa en timer, para el actionTimer del tipo SessionUnbind, etc. */ = false) throw(anna::RuntimeException); // returns true if done at call time (no pendings or ignore pendings, except Disconnecting state by mean DPR/DPA)


  /**
     Class string representation
     \return String with relevant information for this instance.
  */
  /* virtual */std::string asString() const throw();


  /**
     Class xml representation
     \param parent Parent XML node on which hold this instance information.
     \return XML document with relevant information for this instance.
  */
  /* virtual */anna::xml::Node* asXML(anna::xml::Node* parent) const throw();

protected:

  // Deprecated state
  bool a_deprecated;

private:

  // Receiver factory
  ReceiverFactoryImpl<ServerSession, ServerSessionReceiver> a_receiverFactory;

  // Parent information
  LocalServer *a_parent;

  // Client Socket
  anna::comm::ClientSocket *a_clientSocket;

  // Auxiliary messages:
  Message a_cer, a_dwr;

  /* virtual */void expire(anna::timex::Engine *timeController) throw(anna::RuntimeException);

  // Activity:
  /* virtual */void updateIncomingActivityTime() throw();
  /* virtual */void updateOutgoingActivityTime() throw();
  void countSendings(const diameter::CommandId & cid, unsigned int aid, bool ok) throw();

  // Handlers:
  /**
     Handler about event break connection from diameter client over this server-session.

     When notified, ANNA.diameter.comm generates an diameter::comm::ServerSession::eventResponse for every request with pending answers.
  */
  void eventPeerShutdown() throw();

  /**
     Handler about a request retransmission over the session.

     \param request Message retransmitted
  */
  void eventRequestRetransmission(Message *request) throw();

  /**
     Handler for diameter client responses

     \param response Answer container object for corresponding diameter request
  */
  void eventResponse(const Response& response) throw(anna::RuntimeException);

  /**
     Handler for diameter client requests

     \param request Request data block object for corresponding diameter reception
  */
  void eventRequest(const anna::DataBlock& request) throw(anna::RuntimeException);
  //void eventRequest(const Message& request) throw(anna::RuntimeException);

  /**
     Handler for diameter client responses out of context

     \param response Answer data block object without context match
  */
  void eventUnknownResponse(const anna::DataBlock& response) throw(anna::RuntimeException);

  /**
     Handler for diameter client Disconnect-Peer-Answer messages

     \param response Answer data block object without context match
  */
  void eventDPA(const anna::DataBlock& response) throw(anna::RuntimeException);



  /**
  * Handlers for receptions
  */
  /* virtual */void receive(const anna::comm::Message& message) throw(anna::RuntimeException);
  /* virtual */void finalize() throw();

  /* virtual */void expireResponse(Response*) throw();
  void sendCEA() throw(anna::RuntimeException);
  void sendDWA() throw(anna::RuntimeException);


  friend class anna::diameter::comm::Timer;
  friend class LocalServer;
  friend class Engine;
  friend class ServerSessionReceiver;
};

}
}
}

#endif

