// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_comm_Session_hpp
#define anna_diameter_comm_Session_hpp


// STL
#include <string>

#include <anna/core/util/SortedVector.hpp>
#include <anna/core/util/Millisecond.hpp>
#include <anna/core/RuntimeException.hpp>
#include <anna/timex/Timer.hpp>

#include <anna/diameter/defines.hpp>
#include <anna/diameter.comm/ClassCode.hpp>
#include <anna/diameter.comm/Message.hpp>
#include <anna/diameter.comm/Timer.hpp>


namespace anna {
class DataBlock;
namespace timex {
class Engine;
}
}


namespace anna {

namespace diameter {

namespace stack {
class Engine;
}

namespace comm {

class Timer;
class Engine;
class Response;



/**
   Modela la conexion realizada contra un servidor diameter.
*/
class Session : public anna::timex::Timer {

public:

  Session(const char *className, const char *timerName);
  //virtual ~Session();

  /**
   * Default timeout for application message requests over the session.
   */
  static const anna::Millisecond DefaultTimeout;

  /**
   * Default diameter port
   */
  static const int DefaultPort;




  /**
     Session states
     \see diameter::comm::Session::getState
  */
  struct State {
    enum _v {
      /* client + server */ Closed, /**< Closed */
      /* client          */ WaitingBind, /**< Connection confirmation pending*/
      /* client + server */ Bound, /**< Connection done included level application */
      /* client          */ Failover, /**< Last DWR timed out */
      /*          server */ Suspect, /**< Inactivity detected on session */

// .......


      // Cierre de iniciativa local:
      // 1. Envio DPR al PCRF y me pongo en estado 'WaitingDPA'. En este estado no habr� keep-alive DWR/DWA.
      // 2. No dejo pasar nuevas peticiones (BLOCK-SEND).
      // 3. Cierro al recibir el DPA.
      // 4. Si expira el DPA, tambien cierro.
      WaitingDPA, /**< After requesting DPR to server, send is blocked over the session: when DPA arrives (or answer expires) the session is closed */

      // Cierre de iniciativa remota:
      // 1. Recibo DPR del PCRF y me pongo en estado 'Disconnecting'. En este estado no habr� keep-alive DWR/DWA.
      // 2. No dejo pasar nuevas peticiones (BLOCK-SEND).
      // 3. Espero cursar las peticiones pendientes (a m�s tardar, ser� una expiracion Tx desde la recepcion del DPR).
      // 4. Envio DPA y activo un temporizador de cierre local (2*Tx) como proteccion (por si el servidor no cierra).
      Disconnecting, /**< After receiving DPR from server, send is blocked over the session: when no pending requests, DPA is sent to the server who will close connection */

      // Cierre abrupto con/sin espera de pendings:
      Closing /** Planned local disconnection without DPR; send is blocked over the session. Immediate unbind is used when IgnorePendings behaviour is configured */
    };
  };

  /**
     Session state.
     \return State for this session.
  */
  State::_v getState() const throw() { return a_state; }

  /**
   * Defines behaviour on 'Disconnecting' state, about #unbind action (ignore/wait pending answers)
   */
  struct OnDisconnect { enum _v { IgnorePendings, WaitPendings /* graceful unbind */ }; };

  /**
   * Sets behaviour on 'Disconnecting' state, about #unbind action (ignore/wait pending answers).
   * \param onDisconnect Behaviour on 'Disconnecting' state, about #unbind action (ignore/wait pending answers).
   */
  void setOnDisconnect(const OnDisconnect::_v onDisconnect)  throw() { a_onDisconnect = onDisconnect; }

  /**
   * Returns behaviour on 'Disconnecting' state, about #unbind action (ignore/wait pending answers).
   * \return behaviour on 'Disconnecting' state, about #unbind action (ignore/wait pending answers).
   */
  OnDisconnect::_v getOnDisconnect() const throw() { return a_onDisconnect; }



  /**
     Diameter server address, ip or hostname (remote for client-session, local for server-session).
     \return Diameter server address (remote for client-session, local for server-session).
  */
  virtual const std::string& getAddress() const throw() = 0;

  /**
     Diameter server listen port (remote for client-session, local for server-session).
     \return Diameter server listen port (remote for client-session, local for server-session).
  */
  virtual int getPort() const throw() = 0;

  /**
     Socket id.
     \return Socket id.
  */
  int getSocketId() const throw() { return a_socketId; }

  /**
     Returns the next hop-by-hop which will be used over the diameter session to send a message
     It is recommended to fix the message with this value (or store along with the message),
     for application context identification purposes
  */
  const HopByHop & getNextHopByHop() const throw() { return a_nextHopByHop; }

  /** Returns the next end-to-end which will be used over the diameter session to send a message
      It is recommended to fix the message with this value (or store along with the message),
      for application context identification purposes
  */
  const EndToEnd & getNextEndToEnd() const throw() { return a_nextEndToEnd; }


  /**
     Set timeout to consider failed a request.
     \param v Requests class code.
     \param millisecond Milliseconds wait before considering the requests failed.

     Timers are internally managed and automatically activated.
  */
  void setClassCodeTimeout(const ClassCode::_v v, const anna::Millisecond & millisecond) throw() { a_timeouts [v] = millisecond;  }

  /**
   * Timeout configured for class code \em v requests.
   * \return Timeout configured for class code \em v requests.
   */
  anna::Millisecond getClassCodeTimeout(const ClassCode::_v v) const throw() { return a_timeouts [v]; }


  /**
     Returns \em true when diameter server is connected (application level) \em false in other case.
     \return \em true when diameter server is connected (application level) \em false in other case.
  */
  bool isBound() const throw() { return a_state == State::Bound; }


// Envia el mensaje recibido como parametro al servidor con el que estamos conectados mediante esta
// sesion diameter. Dicho mensaje puede ser una peticion o una respuesta (no temporizada).
//
// En caso de enviar una peticion se activara automaticamente un temporizador. Si este llegara a caducar
// se cancelara la busqueda y se invocara al metodo Session::eventResponse indicado que se ha producido
// un error de temporizaci�n. La duracion del temporizador sera la establecida por
// diameter::comm::TimerManager::setTimeout o el valor defecto.
//
// \param message Mensaje a enviar al servidor diameter con el que estamos conectados.
// @return Diameter response reference asociated to a request. NULL if answer sent.
// \warning Solo se podra hacer uso de este metodo cuando el metodo #isBound devuelva \em true.
  virtual const Response* send(const Message* message) throw(anna::RuntimeException) = 0;
  const Response* send(const Message& message) throw(anna::RuntimeException) { return send(&message); }

// Desconecta del extremo remoto
// Se notifica la terminaci�n de cada una de las peticiones pendientes invocando al m�todo Session::eventResponse
// \warning Despu�s de invocar a este m�todo habr�a que volver a iniciar una sesion.
  virtual bool unbind(bool forceDisconnect /* se usa en timer, para el actionTimer del tipo SessionUnbind, etc. */ = false) throw(anna::RuntimeException) = 0;
  // returns true if done at call time (no pendings or ignore pendings, except Disconnecting state by mean DPR/DPA)


  /**
     Gets the timestamp for last incoming activity over the session.

     @return Last incoming activity timestamp.
  */
  const anna::Millisecond & getLastIncomingActivityTime() const throw() { return a_lastIncomingActivityTime; }

  /**
     Gets the timestamp for last outgoing activity over the session.

     @return Last outgoing activity timestamp.
  */
  const anna::Millisecond & getLastOutgoingActivityTime() const throw() { return a_lastOutgoingActivityTime; }

  /**
     Gets the number of requests messages over-the-air.

     @return OTA messages.
  */
  int getOTARequests() const throw() { return a_responses.size(); }


  /**
     Returns idle state (no pending answers).

     @return Idle state.
  */
  bool idle() const throw() { return (getOTARequests() == 0); }


  /**
     In order to avoid message burst during failover procedures (alternate server forwardings done by application),
     orphan request notification could be deferred to expiration event. This is the default behaviour, however if
     need to notice quickly the transport failure for such requets, 'false' should be established. Result code of
     'OrphanTimeout' will be configured at response instance.

     This only applies to client-sessions because is not usual to send request from server sessions and there wont
     be impact notifying orphan requests at the moment of the transport failure.

     @defer Delayed notification for orphan request due to transport failures
  */
  void notifyOrphansOnExpiration(bool defer = true) throw() { a_notifyOrphansOnExpiration = defer; }

  /**
     Class string representation
     \return String with relevant information for this instance.
  */
  virtual std::string asString() const throw();


  /**
     Class xml representation
     \param parent Parent XML node on which hold this instance information.
     \return XML document with relevant information for this instance.
  */
  virtual anna::xml::Node* asXML(anna::xml::Node* parent) const throw();

protected:

  // Auxiliary messages:
  Message a_dpr;

  // Internal, traces, etc.
  const char *a_className;

  // Main session attributes
  int a_socketId; // multiple connection functionality
  State::_v a_state;
  OnDisconnect::_v a_onDisconnect;
  Engine *a_engine;
  anna::diameter::comm::Timer *a_actionTimer;

  // Sequencing
  HopByHop a_nextHopByHop;
  EndToEnd a_nextEndToEnd;
  virtual void initialize() throw();
  void initializeSequences() throw(); // debe invocarse despues de haber asignado el a_parent
  void generateNextSequences() throw() { a_nextHopByHop++; a_nextEndToEnd++; }

  // Context Responses
  struct SortById {
    static HopByHop value(const Response*) throw();
  };
  typedef anna::SortedVector <Response, SortById, HopByHop> response_container;
  typedef response_container::iterator response_iterator;
  typedef response_container::const_iterator const_response_iterator;
  response_container a_responses;
  bool a_notifyOrphansOnExpiration;

  void response_add(Response* response) throw();
  void response_erase(Response* response) throw();
  Response* response_find(const HopByHop hopByHop) throw(anna::RuntimeException);

  response_iterator response_begin() throw() { return a_responses.begin(); }
  response_iterator response_end() throw() { return a_responses.end(); }
  static Response* response(response_iterator ii) throw() { return response_container::data(ii); }

  const_response_iterator response_begin() const throw() { return a_responses.begin(); }
  const_response_iterator response_end() const throw() { return a_responses.end(); }
  static const Response* response(const_response_iterator ii) throw() { return response_container::data(ii); }

  // Activity
  anna::timex::Engine* a_timeController;
  anna::Millisecond a_lastIncomingActivityTime;   // last unix timestamp (in milliseconds) when message reception was managed over the session
  anna::Millisecond a_lastOutgoingActivityTime;   // last unix timestamp (in milliseconds) when message sending was managed over the session
  virtual void updateIncomingActivityTime() throw();
  virtual void updateOutgoingActivityTime() throw();

  // Self-timer expiration handler
  virtual void expire(anna::timex::Engine *timeController) throw(anna::RuntimeException) {;}

  // Timming:
  anna::Millisecond a_timeouts [ClassCode::Max];

  // Handlers:
  /**
     Handler about event break connection over the session.

     When notified, ANNA.diameter.comm generates an diameter::comm::ClientSession::eventResponse for every request with pending answers.
  */
  virtual void eventPeerShutdown() throw() = 0;

  /**
     Handler about a request retransmission over the session.

     \param request Message retransmitted
  */
  virtual void eventRequestRetransmission(Message *request) throw() = 0;

  /**
     Handler for diameter session responses

     \param response Answer data block object for corresponding diameter request
  */
  virtual void eventResponse(const Response& response) throw(anna::RuntimeException) = 0;

  /**
     Handler for diameter session requests

     \param request Request container object for corresponding diameter reception
  */
  virtual void eventRequest(const anna::DataBlock& request) throw(anna::RuntimeException) = 0;
  //void eventRequest(const Message& request) throw(anna::RuntimeException);


  /**
     Handler for diameter session responses out of context

     \param response Answer data block object without context match
  */
  virtual void eventUnknownResponse(const anna::DataBlock& response) throw(anna::RuntimeException) = 0;

  /**
     Handler for diameter session Disconnect-Peer-Answer messages

     \param response Answer data block object without context match
  */
  virtual void eventDPA(const anna::DataBlock& response) throw(anna::RuntimeException) = 0;



  /**
  * Handlers
  */
  virtual void receive(const anna::comm::Message& message) throw(anna::RuntimeException) = 0;
//PROTOCOL ERRORS
//The errors at the protocol level are reported in response messages that contain the �E� bit and the error code in the AVP result-Code (various errors having been produced only the first one of them is reported). Examples of these errors are:
//An unrecognized AVP with the �M� bit is received.
//An AVP is received with an unrecognized value (in the AVP failed-AVP indicates the attribute that the error caused).
//An mandatory AVP is not received.
//Length of operation incorrect.
//Length of AVP incorrect.
//Coding of some AVP incorrect.
//Erroneous inclusion of parameters.





  virtual void finalize() throw(); // invoked from ClientSessionReceiver::eventBreakConnection()


  virtual void expireResponse(Response*) throw();
  void sendDPA() throw(anna::RuntimeException);
  void activateActionTimer(const anna::diameter::comm::Timer::Type::_v type) throw();
  void cancelActionTimer() throw();
  void activateTimer() throw(); // Session timer
  void cancelTimer() throw(); // Session timer
  virtual void timerStopped() throw() {;}
  virtual void timerStarted() throw() {;}


  virtual void setState(State::_v state) throw();

  //anna::diameter::comm::Timer *getActionTimer() const { return (a_actionTimer); }


  // helpers
  static const char* asText(const State::_v) throw();
  static const char* asText(const OnDisconnect::_v) throw();


  friend class anna::diameter::comm::Timer;
  friend class Engine;
  friend class Response;
};

}
}
}

#endif

