// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_http_wims20_ClientSide_hpp
#define anna_http_wims20_ClientSide_hpp

#include <anna/core/functions.hpp>

#include <anna/http/wims20/Abstract.hpp>

namespace anna {

namespace http {

class Request;

namespace wims20 {

/**
   Permite interpretar una URI segn las recomendaciones de WIMS 2.0, lo que facilita
   el desarrollo de aplicaciones integradas en Web 2.0; estas recomendaciones indican
   cmo debe formarse la peticin ClientSide (REpresentational State Transfer) para permitir
   el desarrollo de cualquier servicio.

   Implementa el interfaz WIMS 2.0 desde el punto de vista del proceso que genera la peticin.

   El formato general de una URI segn la recomendacin de WIMS 2.0 es:

   <p>
http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters
   </p>

   Dnde los campos tienen siguen la siguiente especificacin:
   \li http://domain-openapis: Identifica el recurso del Open API.
   \li path-openapis: Recurso opcional que ajusta la ruta hacia los recursos de ste API.
   \li serviceID: Identificador de recurso.
   \li guid: Identificador del usuario que solicita la peticin.
   \li other_possible_level: Opcionalmente se pueden indicar tantos niveles jerrquicos como fuera
   necesario para el servicio.
   \li query_parameters: Lista de parmetros. Si hay ms de un parmetro se separar con '&'.
*/
class ClientSide : public Abstract {
public:
  /**
   * Contructor indicando el parmetro opcional \em path-openapis. Estos dos parmetros se obtendr como
   * parte de la configuracin de nuestro sistema.
   * \param domain: Identifica el recurso del OpenAPI.
   * \param path: Parmetro opcional que ajusta la ruta hacia los recusos de ste API.
   */
  ClientSide(const std::string& domain, const std::string& path) :
    Abstract("ClientSide", domain, path) {
    a_strOtherLevels = NULL;
  }

  /**
   * Constructor que no usar el parmetro opcional \em path-openapis. Este parmetro se obtendr como
   * parte de la configuracin de nuestro sistema.
   * \param domain: Identifica el recurso del OpenAPI.
   */
  explicit ClientSide(const std::string& domain) :
    Abstract("ClientSide", domain) {
    a_strOtherLevels = NULL;
  }

  /**
   * Destructor.
   */
  ~ClientSide() { Abstract::destroyString(a_strOtherLevels); }

  /**
   * Ampla el campo \em other_possible_level con el valor recibido.
   * \param otherLevel Valor con el que ampliar la ruta.
   */
  void addOtherLevel(const std::string& otherLevel) throw(RuntimeException);

  /**
   * Ampla el campo \em other_possible_level con el valor recibido.
   * \param otherLevel Valor con el que ampliar la ruta.
   */
  void addOtherLevel(const char* otherLevel) throw(RuntimeException) {
    std::string aux(otherLevel);
    addOtherLevel(aux);
  }

  /**
   * Establece el valor del parmetro indicado como parmetro. Si el parmetro indicado ya
   * existe en la lista de parmetros registrados su valor se sobreescribe.
   * \param parameter Nombre del parmetro a establecer.
   * \param value Valor asociado al parmetro.
   */
  void setParameter(const char* parameter, const char* value) throw(RuntimeException) {
    std::string p(parameter);
    std::string v(value);
    Abstract::parameter_set(p, v);
  }

  /**
   * Establece el valor del parmetro indicado como parmetro. Si el parmetro indicado ya
   * existe en la lista de parmetros registrados su valor se sobreescribe.
   * \param parameter Nombre del parmetro a establecer.
   * \param value Valor asociado al parmetro.
   */
  void setParameter(const char* parameter, const int value) throw(RuntimeException) {
    std::string p(parameter);
    Abstract::parameter_set(p, anna::functions::asString(value));
  }

  /**
   * Establece el valor del parmetro indicado como parmetro. Si el parmetro indicado ya
   * existe en la lista de parmetros registrados su valor se sobreescribe.
   * \param parameter Nombre del parmetro a establecer.
   * \param value Valor asociado al parmetro.
   */
  void setParameter(const std::string& parameter, const std::string& value) throw(RuntimeException) {
    Abstract::parameter_set(parameter, value);
  }

  /**
   * Establece el valor del parmetro indicado como parmetro. Si el parmetro indicado ya
   * existe en la lista de parmetros registrados su valor se sobreescribe.
   * \param parameter Nombre del parmetro a establecer.
   * \param value Valor asociado al parmetro.
   */
  void setParameter(const std::string& parameter, const int value) throw(RuntimeException) {
    Abstract::parameter_set(parameter, anna::functions::asString(value));
  }

  /**
   * Limpia el contenido asociado al parmetro \em other_possible_level. Slo debera
   * invocarse a este mtodo en caso de que el servicio destino de la peticin haya cambiado.
   */
  void clearOtherLevels() throw() {
    Abstract::clearOtherLevels();
    Abstract::destroyString(a_strOtherLevels);
    a_strOtherLevels = NULL;
  }

  /**
   * Recopila la informacin contenida en la peticin REST de WIMS2.0 y la
   * codifica sobre la URI de la peticin HTTP recibida como parmetro.
   * \param message Mensaje HTTP sobre el que se establecer la URI necesaria para
   * realizar la peticion WIMS 2.0.
   */
  void codeOn(http::Request& message) throw(RuntimeException);

private:
  std::string* a_strOtherLevels;
  std::string a_uri;
};

}
}
}

#endif
