// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_time_Date_hpp
#define anna_time_Date_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>

// Standard
#include <time.h>

// Local
#include <anna/time/internal/Timezone.hpp>


namespace anna {
namespace xml {
class Node;
}
}

//------------------------------------------------------------------------------
//---------------------------------------------------------------------- #define
//------------------------------------------------------------------------------

// Unix time reference:  01/01/1970 00:00:00 UTC
// NTP time reference:   01/01/1900 00:00:00 UTC
// -> 70 years have 2207520000 seconds.
// -> Between 1900 and 1970 there are 17 leap years: 1468800 seconds more (17 days)
// TOTAL DIFFERENCE = 2207520000 + 1468800 = 2208988800
#define TIMESTAMP_OFFSET_NTP1900_OVER_UNIX1970   2208988800U
//#define _2K38_EFFECT_LIMIT                       "20380119031407" // UTC


namespace anna {

namespace time {


/**
* Absolute time representation structs class manager.
* Allow any time assignment/extraction regarding any timezone.
* Uses shell TZ assignment or rely on system timezone configuration (in ubuntu, /etc/timezone
* stores the abbreviation and /etc/localtime keeps the timezone file from /usr/share/zoneinfo).
*
* The internal unix timestamp reamain constant and thedate representation depends on the
* configured timezone. Some store methods could specify a certain origin timezone, but the
* instance inner timezone won't be changed until the user want to do that.
*
* <pre>
*
*  Sample code:
*
*  anna::time::functions::initialize();
*  std::cout << "SystemTimezone: " << anna::time::functions::getSystemTimezone().asString() << std::endl;
*
*  anna::time::Date myBirth ("CET");
*  myBirth.store("19741219111500", "EET");
*  std::cout << "myBirth: " << myBirth.asString() << std::endl;
*
*  anna::time::Date same_moment_canary_island("GMT");
*  same_moment_canary_island.store(myBirth.getUnixTimestamp());
*  std::cout << "same_moment_canary_island: " << same_moment_canary_island.asString() << std::endl;
*  myBirth.setTz("GMT");
*  std::cout << "myBirth on GMT: " << myBirth.asString() << std::endl;
*  std::cout << "EQUAL: same_moment_canary_island = " << same_moment_canary_island.yyyymmddHHmmss() << "; myBirth in context GMT = " << myBirth.yyyymmddHHmmss() << std::endl;
*  same_moment_canary_island.setTz("CET");
*  std::cout << "same_moment_canary_island on CET: " << same_moment_canary_island.asString() << std::endl;
*  anna::time::Date birthday("EET");
*  birthday.store(same_moment_canary_island.getTm(), "CET");
*  std::cout << "EQUAL: birthday TS = " << birthday.getUnixTimestamp() << "; myBirth TS = " << myBirth.getUnixTimestamp() << std::endl;
*  std::cout << "birthday EET = " << birthday.asString() << std::endl;
*
*  std::cout << "Setting Local timezone ..." << std::endl;
*  myBirth.setSystemTimezone();
*  birthday.setSystemTimezone();
*  std::cout << "EQUAL: birthday " << birthday.asString() << "; myBirth " << myBirth.asString() << std::endl;
*  std::cout << "EQUAL: birthday (Local Timezone)= " << birthday.yyyymmddHHmmss() << "; myBirth (Local Timezone)= " << myBirth.yyyymmddHHmmss() << std::endl;
*  std::cout << "myBirth TZ  = " << myBirth.getTzAsString() << std::endl;
*  std::cout << "birthday TZ  = " << myBirth.getTzAsString() << std::endl;
*
*  myBirth.setTz("GMT");
*  std::cout << "myBirth in GMT = " << myBirth.yyyymmddHHmmss() << std::endl;
*
*  Program output:
*
*  SystemTimezone: <TZ unset\>
*  myBirth: Thursday 19/12/1974 10:15:00 CET, isdst = 0 [Unix Timestamp: 156676500, Ntp Timestamp: 2365665300], System Timezone: <TZ unset\>
*  same_moment_canary_island: Thursday 19/12/1974 09:15:00 GMT, isdst = 0 [Unix Timestamp: 156676500, Ntp Timestamp: 2365665300], System Timezone: <TZ unset\>
*  myBirth on GMT: Thursday 19/12/1974 09:15:00 GMT, isdst = 0 [Unix Timestamp: 156676500, Ntp Timestamp: 2365665300], System Timezone: <TZ unset\>
*  EQUAL: same_moment_canary_island = 19741219091500; myBirth in context GMT = 19741219091500
*  same_moment_canary_island on CET: Thursday 19/12/1974 10:15:00 CET, isdst = 0 [Unix Timestamp: 156676500, Ntp Timestamp: 2365665300], System Timezone: <TZ unset\>
*  EQUAL: birthday TS = 156676500; myBirth TS = 156676500
*  birthday EET = Thursday 19/12/1974 11:15:00 EET, isdst = 0 [Unix Timestamp: 156676500, Ntp Timestamp: 2365665300], System Timezone: <TZ unset\>
*  Setting Local timezone ...
*  EQUAL: birthday Thursday 19/12/1974 10:15:00 <TZ unset\>, isdst = 0 [Unix Timestamp: 156676500, Ntp Timestamp: 2365665300], System Timezone: <TZ unset\>; myBirth Thursday 19/12/1974 10:15:00 <TZ unset\>, isdst = 0 [Unix Timestamp: 156676500, Ntp Timestamp: 2365665300], System Timezone: <TZ unset\>
*  EQUAL: birthday (Local Timezone)= 19741219101500; myBirth (Local Timezone)= 19741219101500
*  myBirth TZ  = <TZ unset\>
*  birthday TZ  = <TZ unset\>
*  myBirth in GMT = 19741219091500
*
* </pre>
*/

class Date {
  /////////////////////////////////////////////////
  // Key: Keep coherent all time representations //
  /////////////////////////////////////////////////

  // Main data:
  Timezone a_tz;            // timezone for this instance context
  time_t a_timestamp; // unix timestamp

  // Secondary data:
  struct tm a_tm;


  // sets the current date/time (time(NULL)) and provided TimeZone as described in constructor
  void initialize(const char *TZ = NULL) throw();

  // main refresh method regarding timestamp
  void refresh(void) throw();

  anna::Mutex a_mutex;

public:

  /**
  * Default Constructor
  * @param TZ timezone for date/time representation. Default (NULL) assumes the system
  * timezone configuration. Empty string will be interpreted as UTC timezone.
  */
  Date(const char *TZ = NULL);

  /**
  * Copy Constructor
  */
  Date(const Date & d) { *this = d; }

  /**
  * Destructor
  */
  ~Date() {;}


  // setters

  /**
  * Sets the current date/time (time(NULL)) and provided TimeZone as described in constructor
  */
  void setNow(const char *TZ = NULL) throw() { initialize(TZ); }


  /**
  * Sets the provided timezone for date/time representation of the class instance.
  * This method keeps invariant the unix timestamp, thus, the date representation
  *  will evolve depending on the configured timezone.
  * @warning Be careful with incorrect TZ values because no error will be shown but
  * dates could be misleading. The valid TZ could be seen with tzselect tool or
  * directly looking at /usr/share/zoneinfo hierarchy.
  *
  * @param TZ timezone for date/time representation.
  * Default (NULL) is the local host (user-supplied) timezone
  * Empty string sets UTC0 (GMT, Greenwich)
  * Use available timezones in /usr/share/zoneinfo; i.e.: Europe/Madrid, CET, etc.
  * You can also use 'tzselect' helper, but that tool don't change the system
  * timezone, only helps to know the valid TZ values and shows how to change
  * the timezone settings if you would want to do that).
  */
  void setTz(const char *TZ = NULL) throw();


  /**
   Sets the local host timezone (even if it is based on TZ or in /usr/share/zoneinfo)
  */
  void setSystemTimezone() throw();


  /**
  * Sets date/time providing unix timestamp (seconds since 01-Jan-1970 GMT)
  *
  * @param unixTimestamp Time since 01-Jan-1970 GMT
  */
  void store(const time_t & unixTimestamp) throw();
  void storeUnix(const time_t & unixTimestamp) throw() { store(unixTimestamp); }


  /**
  * Sets date/time providing ntp timestamp (seconds since 01-Jan-1900 GMT)
  *
  * @param ntpTimestamp Time since 01-Jan-1900 GMT
  */
  void storeNtp(const unsigned int &ntpTimestamp) throw();


  /**
  * Sets date/time from standard 'tm' struct and optional timezone
  *
  * @param date 'tm' struct date/time
  * @param TZ timezone for date/time provided. Default (NULL) assumes the system
  * timezone configuration. Empty string will be interpreted as UTC timezone.
  */
  void store(const struct tm &date, const char *TZ = NULL) throw(anna::RuntimeException);


  /**
  * Sets date/time from string representation with certain provided format,
  * and optional timezone
  *
  * @param stringDate Date/time string
  * @param TZ timezone for date/time provided. Default (NULL) assumes the system
  * timezone configuration. Empty string will be interpreted as UTC timezone.
  * @param strptimeFormat Date/time string format for strptime primitive.
  * See format syntax at http://man7.org/linux/man-pages/man3/strptime.3.html
  */
  void store(const std::string &stringDate, const char *TZ = NULL, const char *strptimeFormat = "%Y%m%d%H%M%S") throw(anna::RuntimeException);
  //void store(const std::string &stringDate, const char *tz = NULL, const char *strptimeFormat = "%Y-%m-%d-%H-%M-%S") throw(anna::RuntimeException);


  /**
  * Copy Operator
  *
  * @param d Source class instance
  *
  * @return Returns copied reference
  */
  Date & operator = (const Date &d);

  // gets

  // Operators

  /**
  * Operator ==
  *
  * @param d1 First instance from Date class
  * @param d2 Second instance from Date class
  *
  * @return Returns d1 == d2 comparison
  */
  friend bool operator == (const Date & d1, const Date & d2) {
    return (d1.getUnixTimestamp() == d2.getUnixTimestamp());
  }

  /**
  * Operator !=
  *
  * @param d1 First instance from Date class
  * @param d2 Second instance from Date class
  *
  * @return Returns d1 != d2 comparison
  */
  friend bool operator != (const Date & d1, const Date & d2) {
    return (d1.getUnixTimestamp() != d2.getUnixTimestamp());
  }

  /**
  * Operator <=
  *
  * @param d1 First instance from Date class
  * @param d2 Second instance from Date class
  *
  * @return Returns d1 <= d2 comparison (compare unix timestamp unix, ignore TZ)
  */
  friend bool operator <= (const Date & d1, const Date & d2) {
    return (d1.getUnixTimestamp() <= d2.getUnixTimestamp());
  }

  /**
  * Operator >=
  *
  * @param d1 First instance from Date class
  * @param d2 Second instance from Date class
  *
  * @return Returns d1 >= d2 comparison (compare unix timestamp unix, ignore TZ)
  */
  friend bool operator >= (const Date & d1, const Date & d2) {
    return (d1.getUnixTimestamp() >= d2.getUnixTimestamp());
  }


  /**
  * Gets context instance timezone information
  *
  * @return  NULL if TZ is unset, or string for TZ environment variable
  * (empty is allowed and use to be understood as UTC.
  */
  const char *getTz(void) const throw() { return (a_tz.unsetTZ() ? NULL:a_tz.getValue().c_str()); }


  /**
  * Gets the context timezone as string
  */
  std::string getTzAsString() const throw() { return a_tz.asString(); }


  /**
  * Gets 'tm' struct on the current timezone
  *
  * @return 'tm' struct on the current timezone
  */
  const struct tm & getTm(void) const throw() { return a_tm; }

  /**
  * Gets the week day abbreviation: Sun, Mon, Tue, Wed, Thu, Fri, Sat
  *
  * @return Day of the week
  */
  const char *getDay(void) const throw();


  /**
  * Gets 'yyyymmddHHmmss' representation for the current timezone
  *
  * @return Time/date on format 'yyyymmddHHmmss'
  *
  * @see getTz
  */
  std::string yyyymmddHHmmss(void) const throw();


  /**
  * Gets unix timestamp (01-Jan-1970 GMT)
  *
  * @return Seconds since 01-Jan-1970 GMT
  */
  const time_t & getUnixTimestamp(void) const throw() { return a_timestamp; }


  /**
  * Gets ntp timestamp (01-Jan-1900 GMT)
  *
  * @return Seconds since 01-Jan-1900 GMT
  */
  unsigned int getNtpTimestamp(void) const throw() {
    unsigned int ntp_timestamp = a_timestamp + TIMESTAMP_OFFSET_NTP1900_OVER_UNIX1970;
    return ntp_timestamp;
  }


  /**
  * Class string representation
  *
  * @return String with class content
  */
  std::string asString(void) const throw();


  /**
  * Class XML representation
  *
  * @return XML with class content
  */
  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();

};

}
}

#endif

