// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <sys/ioctl.h>
#include <sys/time.h>

#include <poll.h>

//#include <asm/ioctls.h>

#include <algorithm>

#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/functions.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/app/functions.hpp>

#include <anna/comm/ClientSocket.hpp>
#include <anna/comm/INetAddress.hpp>
#include <anna/comm/TransportFactory.hpp>
#include <anna/comm/Transport.hpp>
#include <anna/comm/Message.hpp>
#include <anna/comm/Network.hpp>
#include <anna/comm/Host.hpp>
#include <anna/comm/Server.hpp>
#include <anna/comm/Communicator.hpp>
#include <anna/comm/ReceiverFactory.hpp>
#include <anna/comm/Receiver.hpp>
#include <anna/comm/CongestionController.hpp>

#include <anna/comm/internal/Poll.hpp>

using namespace std;
using namespace anna;

// static
const Millisecond comm::ClientSocket::DefaultMaxConnectionDelay(200);
const Millisecond comm::ClientSocket::DefaultMaxWriteDelay(500);


void comm::ClientSocket::initialize()
throw() {
   a_status = Status::None;
   a_transport = NULL;
   a_expectedSize = -1;
   a_offset = 0;
   a_cachedServer = NULL;
   a_msMaxConnectionDelay = DefaultMaxConnectionDelay;
   a_msMaxWriteDelay = DefaultMaxWriteDelay;
   a_rcvBufferSize = -1;
   a_receiver = NULL;
}

/*
 * Se invoca desde el handler::ServerSocket y no hace falta protegerlo con una SCC
 * porque ya est protegido por ste.
 */
void comm::ClientSocket::setfd(const int fd)
throw(RuntimeException) {
   Socket::a_fd = fd;
   Socket::a_isBound = true;
   activate(Status::Connected);
   a_offset = 0;

   if (Socket::a_type == Socket::Type::Stream) {
      sockaddr_in addr;
      socklen_t len(sizeof(addr));
      Network& network(Network::instantiate());
      anna_comm_socket_check(
         getsockname(fd, (struct sockaddr *) &addr, &len),
         "Cannot obtain socket name"
      );
      Socket::a_localAccessPoint = INetAddress(network.find(addr.sin_addr.s_addr), ntohs(addr.sin_port));
      anna_comm_socket_check(
         getpeername(fd, (struct sockaddr *) &addr, &len),
         "Cannot obtain peer name"
      );
      a_remoteAccessPoint = INetAddress(network.find(addr.sin_addr.s_addr), ntohs(addr.sin_port));
   }

   setBlockingMode(false);
   getSocketOptions();
   LOGDEBUG(
      string msg("comm::ClientSocket::setfd | ");
      msg += asString();
      Logger::debug(msg, ANNA_FILE_LOCATION);
   );
}

/*
 * Si se establece el receptor directamente, sin tener una factoria intermedia, hay que
 * invocar al metodo comm::Receiver::initialize.
 */
void comm::ClientSocket::setReceiver(comm::Receiver* receiver)
throw(RuntimeException) {
   if (a_receiver && Socket::a_receiverFactory)
      Socket::a_receiverFactory->release(a_receiver);

   a_receiver = receiver;

   if (a_receiverFactory == NULL)
      a_receiver->initialize();
}

// Protegido desde el Socket::close
void comm::ClientSocket::do_close()
throw() {
   Socket::do_close();

   if (a_transport != NULL) {
      a_transport->clear();
      Guard guard(Socket::a_transportFactory, typeid(Socket::a_transportFactory).name());
      Socket::a_transportFactory->release(a_transport);
      a_transport = NULL;
   }

   a_status = Status::None;
   a_expectedSize = -1;
   a_data.clear();
   a_cachedServer = NULL;
   a_offset = 0;

   if (a_receiver) {
      if (Socket::a_receiverFactory)
         Socket::a_receiverFactory->release(a_receiver);

      a_receiver = NULL;
   }
}

comm::Server* comm::ClientSocket::getServer()
throw(RuntimeException) {
   Guard guard(*this, "comm::ClientSocket (getServer)");
   const INetAddress& remoteAddress = getRemoteAccessPoint().getINetAddress();
   const Device* device = remoteAddress.getDevice(false);
   const int remotePort = remoteAddress.getPort();

   if (device == NULL)
      return NULL;

   if (a_cachedServer != NULL)
      if (a_cachedServer->getRemotePort() == remotePort && a_cachedServer->getHost()->contains(device))
         return a_cachedServer;

   Network& network = Network::instantiate();
   comm::Server* result = NULL;
   comm::Server* aux;

   for (Network::host_iterator ii = network.host_begin(), maxii = network.host_end(); ii != maxii; ii ++) {
      if (Network::host(ii)->contains(device) == false)
         continue;

      if ((aux = Network::host(ii)->find_server(remotePort)) == NULL)
         continue;

      result = aux;
      break;
   }

   return (a_cachedServer = result);
}

//_________________________________________________________________________________
// (1) Si termina el periodo de espera sin devolver ninguna actividad => podemos considerar que
// se ha conectado. Es posible que haya algn componente hardware que no permita hacer
// la conexin.
//_________________________________________________________________________________
void comm::ClientSocket::connect()
throw(RuntimeException) {
   Guard guard(*this, "comm::ClientSocket (connect)");
   anna_socket_assert(isConnected() == true, "Already connected");

   if (Socket::isOpened() == false) {
      Socket::open();
      getSocketOptions();
      setBlockingMode(false);
   }

   if (Socket::a_localAccessPoint.isNull() == false && Socket::isBound() == false)
      Socket::bind();

   sockaddr* s(NULL);
   int len(0);
   a_remoteAccessPoint.translate(*this, s, len);
   a_cachedServer = NULL;
   const Millisecond msinit = functions::millisecond();
   Millisecond msnow;
   int xerrno;
   int r;
   int tryCounter(0);
   pollfd waiting;
   msnow = msinit;
   waiting.fd = Socket::a_fd;
   waiting.events = POLLOUT;

   while ((msnow - msinit) <= a_msMaxConnectionDelay) {
      tryCounter ++;

      try {
         if ((r = do_connect(s, len)) != -1)
            break;
      } catch (RuntimeException&) {
         close();
         throw;
      }

      if ((xerrno = errno) == EISCONN) {
         r = 0;
         break;
      }

      if (xerrno != EINPROGRESS && xerrno != EALREADY)
         break;

      anna_signal_shield(r, poll(&waiting, 1, a_msMaxConnectionDelay));

      if (r == 0) {                                                                           // (1)
         r = -1;
         xerrno = ETIMEDOUT;
      }

      msnow = functions::millisecond();
   }

   if (r == -1) {
      std::string msg("comm::ClientSocket::connect | ");
      msg += asString();
      msg += " | N-loop: ";
      msg += functions::asString(tryCounter);
      msg += " | Wait: ";
      msg += functions::asString(msnow - msinit);
      msg += " ms";
      close();
      throw RuntimeException(msg, xerrno, ANNA_FILE_LOCATION);
   }

   activate(Status::Connected);
   a_offset = 0;
   LOGDEBUG(
      string msg("comm::ClientSocket::connect | ");
      msg += asString();
      Logger::debug(msg, ANNA_FILE_LOCATION);
   );
}

void comm::ClientSocket::send(comm::Message& message)
throw(RuntimeException) {
   Guard guard(*this, "comm::ClientSocket (send)");
   do_write(getTransport()->code(message));
}

void comm::ClientSocket::send(comm::Message* message)
throw(RuntimeException) {
   if (message == NULL)
      throw RuntimeException("comm::ClientSocket::send | Cannot send a NULL message", ANNA_FILE_LOCATION);

   send(*message);
}

//_______________________________________________________________________
// Si el clientsocket no esta marcado como "en proceso", es decir, todavia estamos
// recogiendo mensajes de el, se cierra directamente, en otro cpero aso se marca
// como pendiente de cerrar y cuando se recogan todos los mensajes que contiene
// su buffer, se cerrar.
//_______________________________________________________________________
void comm::ClientSocket::requestClose()
throw() {
   Communicator* communicator = app::functions::component <Communicator> (ANNA_FILE_LOCATION);
   a_status |= Status::ClosePending;
   // Slo tiene efecto real en modo ST
   communicator->notifyPendingClose();
   /*
      else {
         communicator->detach (this);
      }
   */
}

//---------------------------------------------------------------------------------------
// Cuando el communicator detecta actividad en el socket invoca a este metodo.
// (1) Si devuelve 0 es porque han borrado el extremo remoto.
// (2) Si el buffer intermedio tiene mas de 4 veces el numero de bytes planificado =>
// lo cerramos independientemente del tamanio esperado.
// (3) Si el buffer tiene un numero de bytes mayor al planificiado y todavia no ha
// detectado ningun mensaje => lo cerramos.
//
// Se invoca desde un entorno protegido, por lo que no hace falta SSCC.
//---------------------------------------------------------------------------------------
comm::Socket::Notify::_v comm::ClientSocket::receive()
throw(RuntimeException) {
   Notify::_v result;

   if (a_reader.getMaxSize() == 0) {
      string msg(asString());
      msg += " | I/O buffer length was not established";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
   }

   const char* buffer(a_reader.getData());

   const int r(do_read(buffer, a_reader.getMaxSize()));

   a_buffer.setup(buffer, r);

   // Si ha leido el mximo permitido, entonces obtiene los bytes que quedan por leer en el sistema I/O.
   // Si ha ledo menos del mximo, entonces no queda nada en el sistema I/O
   if (r == a_reader.getMaxSize()) {
      if (ioctl(a_fd, FIONREAD, &a_pendingBytes.bytesToRead) != 0)
         a_pendingBytes.bytesToRead = 0;
   } else
      a_pendingBytes.bytesToRead = 0;

   LOGDEBUG(
      string msg("comm::ClientSocket::receive | ");
      msg += asString();
      msg += " | Pending: ";
      msg += anna::functions::asString(a_data, 24);
      msg += functions::asText(" | Received N-bytes: ", a_buffer.getSize());
      msg += ", Buffer: ";
      msg += anna::functions::asString(a_buffer, 24);
      Logger::debug(msg, ANNA_FILE_LOCATION);
   );

   if (r > 0) {                                                              // (1)
      if (getIgnoreIncomingMessages() == true) {
         app::functions::component <Communicator> (ANNA_FILE_LOCATION)->eventIgnoreBurst(*this, a_buffer);
         forgot();
         LOGDEBUG(
            string msg("comm::ClientSocket::receive | ");
            msg += asString();
            msg += " | Ignoring incoming messages";
            Logger::debug(msg, ANNA_FILE_LOCATION);
         );
         return Notify::None;
      }

      result = Notify::ReceiveData;
      a_data += a_buffer;
      const int overQuotaSize(getTransport()->getOverQuotaSize());

      if (a_expectedSize == -1)
         calculeExpectedSize(a_data);

      if (a_expectedSize == -1 && getBufferSize() > overQuotaSize) {                         // (3)
         string msg(asString());
         msg += functions::asText(" | Closed local point due to excessive memory consumption: N-Bytes: ", r);
         msg += functions::asText("/BufferSize: ", getBufferSize());
         msg += functions::asText("/OverQuotaSize: ", overQuotaSize);
         msg += ')';
         Logger::error(msg, ANNA_FILE_LOCATION);
         app::functions::component <Communicator> (ANNA_FILE_LOCATION)->eventOverQuota(*this);
         a_data.clear();
         a_offset = 0;
         result = Notify::Close;
      } else if (isCorrupt() == true)
         result = Notify::Corrupt;
   } else
      result = Notify::Close;

   return result;
}

comm::Socket::Notify::_v comm::ClientSocket::wait(const Millisecond &timeout, const bool _receive)
throw(RuntimeException) {
   Guard guard(*this, "comm::ClientSocket (wait)");

   if (Socket::isOpened() == false) {
      string msg("comm::ClientSocket::wait | ");
      msg += asString();
      msg += " | Is not opened";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
   }

   Poll poll;
   poll.setTimeout(timeout);
   poll.insert(a_fd);
   poll.waitMessage();
   Notify::_v result = (poll.fetch() != -1) ? ((_receive == true) ? receive() : Notify::ReceiveData) : Notify::None;
   LOGDEBUG(
      string msg("comm::ClientSocket::wait | Timeout: ");
      msg += functions::asString(timeout);
      msg += functions::asText(" ms | Receive: ", _receive);
      msg += functions::asText(" | fd: ", getfd());
      msg += " | Result: ";
      msg += Socket::asText(result);
      Logger::debug(msg, ANNA_FILE_LOCATION);
   );
   return result;
}

//---------------------------------------------------------------------------------------
// Cuando el ClientSocket notifica que ha recibido datos, el communicator invoca a este
// metodo por si hubiera algn mensaje completo.
//
// (1) Si todavia no conoce el tamao esperado del mensaje => intenta calcularlo.
// (2) Si ya conoce la longitud y en el buffer recibido quedan , al menos, ese nmero
// de bytes => hay al menos un mensaje completo. Lo procesa y deja todo preparado
// para la proxima invocacion de este metodo.
// (3) Todavia no tenemos un mensajes completo, lo guardamos y esperamos a que se vuelva
// a leer lo que falta desde 'ClientSocket::receive'.
// (4) Si nos quedan bytes por analizar pero no tenemos suficientes para calcular
// la longitud esperada => prepara el a_data para recibir el siguiente bloque.
//---------------------------------------------------------------------------------------
const DataBlock* comm::ClientSocket::fetch()
throw(RuntimeException) {
   if (a_status & Status::ClosePending)
      return NULL;

   const int remainingSize(a_data.getSize() - a_offset);

   if (remainingSize <= 0) {
      a_data.clear();
      a_offset = 0;
      return NULL;
   }

   const DataBlock* result(NULL);

   a_buffer.setup(a_data.getData() + a_offset, remainingSize);

   if (a_expectedSize == -1)                                               // (1)
      calculeExpectedSize(a_buffer);

   if (isCorrupt() == true)
      return NULL;

   if (a_expectedSize != -1) {
      if (a_expectedSize <= remainingSize) {                               // (2)
         a_buffer.resize(a_expectedSize);
         result = &a_buffer;
         a_offset += a_expectedSize;
         a_expectedSize = -1;
      }

      // (3)
   }

   // (4)

   if (result == NULL) {
      if (a_offset > 0) {
         if (a_offset >= a_data.getSize())
            a_data.clear();
         else
            a_data.remove(a_offset);

         a_offset = 0;
      }

      LOGDEBUG(
         string msg("comm::ClientSocket::fetch | ");
         msg += asString();
         msg += functions::asText(" | N-bytes: ", a_data.getSize());
         msg += " | Pending: ";
         msg += anna::functions::asString(a_data, 24);
         Logger::debug(msg, ANNA_FILE_LOCATION);
      );
   }

   return result;
}

void comm::ClientSocket::calculeExpectedSize(const DataBlock& data)
throw() {
   try {
      if ((a_expectedSize = getTransport()->calculeSize(data)) != -1) {
         LOGDEBUG(
            string msg("comm::ClientSocket::calculeExpectedSize | ");
            msg += asString();
            Logger::debug(msg, ANNA_FILE_LOCATION)
         );
      }

      if (a_expectedSize < -1) {
         activate(Status::Corrupt);
         string msg(asString());
         msg += " | Invalid expected size";
         Logger::error(msg, ANNA_FILE_LOCATION);
      }
   } catch (RuntimeException& ex) {
      ex.trace();
      activate(Status::Corrupt);
   }
}

comm::Transport* comm::ClientSocket::reserveTransport()
throw(RuntimeException) {
   if (a_transport == NULL) {
      if (Socket::a_transportFactory == NULL) {
         string msg(asString());
         msg += " | Transport factory was not especified";
         throw RuntimeException(msg, ANNA_FILE_LOCATION);
      }

      Guard guard(this, "comm::ClientSocket::reserveTransport");

      if (a_transport == NULL) {
         Guard guard(Socket::a_transportFactory, typeid(Socket::a_transportFactory).name());
         a_transport = Socket::a_transportFactory->create();
         a_transport->clear();
      }
   }

   return a_transport;
}

comm::Transport* comm::ClientSocket::unsafe_reserveTransport()
throw(RuntimeException) {
   if (a_transport == NULL) {
      if (Socket::a_transportFactory == NULL) {
         string msg(asString());
         msg += " | Transport factory was not especified";
         throw RuntimeException(msg, ANNA_FILE_LOCATION);
      }

      a_transport = Socket::a_transportFactory->create();
      a_transport->clear();
   }

   return a_transport;
}

comm::Receiver* comm::ClientSocket::reserveReceiver()
throw(RuntimeException) {
   if (a_receiver == NULL && Socket::a_receiverFactory != NULL) {
      Guard guard(this, "comm::ClientSocket::reserveReceiver");

      // NO hace falta proteger porque ya lo hace su 'create'
      if (a_receiver == NULL)
         a_receiver = Socket::a_receiverFactory->create();
   }

   return a_receiver;
}

// Este metodo se sobre-escribira en commsec::ClientSocket para establecer la conexion mediante las
// funciones SSH
int comm::ClientSocket::do_connect(const sockaddr* s, const  int len)
throw(RuntimeException) {
   int r;
   anna_signal_shield(r, ::connect(a_fd, s, len));
   return r;
}

//----------------------------------------------------------------------------------------------------
// Este metodo se sobre-escribe en comm::DatagramSocket
//
// (1) si no ha sido capaz de escribir todo el mensaje tenemos que esperar a que el bloque anterior
// haya salido, ya que si intentamos escribir de forma contienua habra un momento que obtengamos
// un error 'Resource temporarily unavailable'.
// (2) La cola de salida esta llena momentaneamente, esperamos unos milisegundos para volver a
// tratar de enviar.
// (3) Obtenmos el error "Broken Pipe" ya solo se puede cerrar el socket y tratar de comenzar
// (4) Si envi un trozo, pero no pudo enviar el mensaje completo => cierra el socket para evitar
//     problemas de interpretacin en el extremo remoto.
//----------------------------------------------------------------------------------------------------
void comm::ClientSocket::do_write(const DataBlock& message)
throw(RuntimeException) {
   int size = message.getSize();
   const char* data =  message.getData();
   int r(0);
   int nloop(0);
   int cx(0);
   bool retry = true;
   bool isok = false;
   int xerrno;
   bool sendSomething = false;

   // Si ha enviado parte del mensaje => tiene que enviarlo completo
   while (retry == true && cx < 5) {
      anna_signal_shield(r, write(Socket::a_fd, data, size));

      if (r == size) {
         isok = true;
         break;
      }

      xerrno = errno;

      if (r < 0) {
         if (xerrno == EAGAIN) {                              // (2)
            cx ++;
            pollfd waiting;
            waiting.fd = Socket::a_fd;
            waiting.events = POLLOUT;
            anna_signal_shield(r, poll(&waiting, 1, a_msMaxWriteDelay));
         } else {
            // Si hay un error definitivo deja de intentarlo y lanza la excepcin.
            // Si se ha enviado algo supondremos que el extremo remoto es capaz de recuperarse
            retry = false;
         }
      } else  { /* (r < left) */
         nloop ++;
         data += r;
         size -= r;
         cx = 0;
         sendSomething = true;
      }
   }

   if (isok == false) {
      if (xerrno == EPIPE || sendSomething == true) {                               // (3)(4)
//         activate (Status::ClosePending);
         requestClose();
         r = shutdown(getfd(), SHUT_WR);
      }

      string msg("comm::ClientSocket::do_write | fd: ");
      msg += functions::asString(getfd());
      throw RuntimeException(msg, xerrno, ANNA_FILE_LOCATION);
   }

   LOGDEBUG(
      string msg("comm::ClientSocket::do_write | ");
      msg += asString();
      msg += functions::asText(" | N-Loop: ", nloop);
      msg += functions::asText(" | Sent: ", message);
      Logger::debug(msg, ANNA_FILE_LOCATION);
   );
}

//----------------------------------------------------------------------------
// (1) Cuando intentamos leer del socket asociado a un servidor que ha caido
//     no devuelve 0 (como cuando se cae el socket de un cliente) sino que
//     devuelve el error (ECONNRESET = 131).
//
// Devuelve 0 para indicar que el Socket se ha cerrado en el extremo remoto
//----------------------------------------------------------------------------
int comm::ClientSocket::do_read(const char* data, const int maxSize)
throw(RuntimeException) {
   int result;

   do {
      if ((result = ::read(Socket::a_fd, (void*) data, maxSize)) < 0) {
         if (errno == EINTR)
            continue;

         if (errno == ECONNRESET) { // (1)
            result = 0;
            break;
         }

         string msg(asString());
         msg += " | Cannot receive";
         throw RuntimeException(msg, errno, ANNA_FILE_LOCATION);
      }
   } while (errno == EINTR && result < 0);

   return result;
}

void comm::ClientSocket::forgot()
throw() {
   if (a_transport != NULL)
      a_transport->clear();

   deactivate(Status::Corrupt);
   a_expectedSize = -1;
   a_data.clear();
   a_offset = 0;
}

//--------------------------------------------------------------------------------------
// (1) Recupera el tamano para verificar que se ha podido establecer la solicitud.
//--------------------------------------------------------------------------------------
void comm::ClientSocket::getSocketOptions()
throw(RuntimeException) {
   if ((a_rcvBufferSize = comm::Communicator::getReceivingChunkSize()) == -1) {
      socklen_t l = sizeof(int);
      int fd = Socket::getfd();
      anna_comm_socket_check(
         getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &a_rcvBufferSize, &l),
         "Cannot obtain SO_RCVBUF"
      );
   }

   a_reader.allocate(a_rcvBufferSize);
}

int comm::ClientSocket::getTotalPendingBytes() const
throw() {
   // El n de bytes pendientes de procesar ser lo que queda en la cola I/O + los bytes que tenemos cargados en memoria
   // los bytes pendientes se obtiene al leer el chunk
   return a_pendingBytes.bytesToRead + getBufferSize();
}


string comm::ClientSocket::asString() const
throw() {
   string msg("comm::ClientSocket { ");

   if (this == NULL)
      return msg += " <freed> }";

   msg += Socket::asString();
   msg += functions::asText(" | RcvBufferSize: ", a_rcvBufferSize);
   msg += " bytes | Status: ";
   msg += Status::asString(a_status);
   msg += functions::asString("| MaxConDelay: %d ms", a_msMaxConnectionDelay.getValue());

   if (a_transport)
      msg += functions::asText(" | OverQuotaSize: ", a_transport->getOverQuotaSize());

   msg += functions::asString(" | Reserved: %d | Pending: %d | Offset: %d | ExpectedSize: %d", a_data.getSize(), getBufferSize(), a_offset, a_expectedSize);
   msg += " | Remote access point: ";
   a_remoteAccessPoint.asString(msg);
   return msg += " }";
}

xml::Node* comm::ClientSocket::asXML(xml::Node* parent) const
throw(RuntimeException) {
   xml::Node* clientSocket = parent->createChild("comm.ClientSocket");

   if (this == NULL) {
      clientSocket->createAttribute("Freed", "yes");
      return clientSocket;
   }

   Socket::asXML(clientSocket);
   clientSocket->createAttribute("Status", Status::asString(a_status));
   clientSocket->createAttribute("RcvBufferSize", a_rcvBufferSize);
   clientSocket->createAttribute("MaxConnDelay", a_msMaxConnectionDelay);
   clientSocket->createAttribute("IgnoreIncomingMessages", functions::asString(a_ignoreIncomingMessages));

   if (a_transport)
      clientSocket->createAttribute("OverQuotaSize", a_transport->getOverQuotaSize());

   xml::Node* buffer = clientSocket->createChild("Buffer");
   buffer->createAttribute("Reserved", a_data.getSize());
   buffer->createAttribute("Pending", getTotalPendingBytes());
   a_remoteAccessPoint.asXML("comm.RemotePoint", clientSocket);

   if (a_receiver)
      a_receiver->asXML(clientSocket);

   return clientSocket;
}

string comm::ClientSocket::Status::asString(const int status)
throw() {
   string result;

   if (status == 0)
      result = "None ";
   else {
      if (status & Status::Connected)
         result = "Connected ";

      if (status & Status::Corrupt)
         result += "Corrupt ";

      if (status & Status::ClosePending)
         result += "ClosePending ";

      if (status & Status::Working)
         result += "Working ";
   }

   return result;
}


