// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/functions.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>
#include <anna/xml/Text.hpp>

using namespace std;
using namespace anna;
using namespace anna::xml;

/*
 * Si el nodo no tiene padre => debe crear los pool de datos en los que sus descendientes irn
 * creando los objetos que vayan siendo necesarios.
 *
 * ste es el constructor que se usar para instanciar al nodo ROOT.
 */
Node::Node(const char* name) : a_name(name), a_parent(NULL), a_text(NULL), a_namespace(NULL) {
  a_node_pool = new node_pool;
  a_attribute_pool = new attribute_pool;
  a_text_pool = new text_pool;
  a_namespace_pool = new namespace_pool;
  a_namespaces = new namespace_container;
  a_root = this;
}

/*
 * Si el nodo tiene padre => copiar las instancias de los pools de datos.
 */
Node::Node() :  a_parent(NULL), a_text(NULL),  a_namespace(NULL) {
  a_node_pool = NULL;
  a_attribute_pool = NULL;
  a_text_pool = NULL;
  a_namespace_pool = NULL;
  a_namespaces = NULL;
  a_root = NULL;
}

Node::~Node() {
  clear();

  if(a_parent == NULL) {
    delete a_node_pool;
    delete a_attribute_pool;
    delete a_text_pool;
    delete a_namespace_pool;
    delete a_namespaces;
  }
}

void Node::clear()
throw() {
  for(attribute_iterator aa = a_attributes.begin(), maxaa = a_attributes.end(); aa != maxaa; aa ++)
    a_attribute_pool->release(attribute(aa));

  a_attributes.clear();

  for(Children::iterator cc = a_children.begin(), maxcc = a_children.end(); cc != maxcc; cc ++) {
    Node* aux = node(cc);
    aux->clear();
    a_node_pool->release(aux);
  }

  a_children.clear();

  if(a_text != NULL) {
    a_text_pool->release(a_text);
    a_text = NULL;
  }

  // Si es el nodo root ....
  if(a_parent == NULL) {
    a_namespaces->clear();
    a_namespace_pool->clear();
  }

  a_namespace = NULL;
}

const Attribute* Node::getAttribute(const char* name, const bool exceptionWhenNotFound) const
throw(RuntimeException) {
  const Attribute* result(NULL);
  result = find(name, attribute_begin(), attribute_end());

  if(result == NULL && exceptionWhenNotFound == true) {
    string msg = asString();
    msg += " | Attribute: ";
    msg += name;
    msg += " | Not found";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  return result;
}

xml::Attribute* Node::createAttribute(const char* name, const char* value, const Namespace* _namespace)
throw() {
  Attribute* attribute = a_attribute_pool->create();
  attribute->setNode(this);
  attribute->setName(name);
  attribute->setValue(value);
  attribute->setNamespace(_namespace);
  a_attributes.push_back(attribute);
  return attribute;
}

xml::Text* Node::createText(const char* text)
throw(RuntimeException) {
  if(a_text != NULL) {
    string msg = asString();
    msg += " | ";
    msg += a_text->asString();
    msg += " | Node has already text assigned";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  a_text = a_text_pool->create();
  a_text->setNode(this);
  a_text->setValue(text);
  return a_text;
}

Node* Node::createChild(const char* name)
throw() {
  Node* result = a_node_pool->create();
  result->a_parent = this;
  result->setName(name);
  result->a_node_pool = a_node_pool;
  result->a_attribute_pool = a_attribute_pool;
  result->a_text_pool = a_text_pool;
  result->a_root = a_root;
  a_children.push_back(result);
  return result;
}

/*
 * Recordar que todos los namespaces se crean slo en el nodo ROOT.
 */
const Namespace* Node::createNamespace(const std::string& name, const char* reference)
throw(RuntimeException) {
  const Namespace* result = NULL;

  if((result = namespace_find(name, false)) != NULL) {
    if(result->getReference() != reference) {
      string msg(result->asString());
      msg += " | New reference: ";
      msg += reference;
      msg += " | Namespace already defined";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }
  } else {
    if(a_root == this) {
      Namespace* newns = a_namespace_pool->create();
      newns->setName(name);
      newns->setReference(reference);
      a_namespaces->add(newns);
      result = newns;
    } else
      result = a_root->createNamespace(name, reference);
  }

  return result;
}

const Node* Node::find(const char* childName, const bool exceptionWhenNotFound) const
throw(RuntimeException) {
  const Node* child;

  for(Node::Children::const_iterator ii = a_children.begin(), maxii = a_children.end(); ii != maxii; ii ++) {
    child = *ii;

    if(anna_strcmp(child->a_name.c_str(), childName) == 0)
      return child;
  }

  if(exceptionWhenNotFound == true) {
    std::string msg("Successor '");
    msg += childName;
    msg += "' not found at node '";
    msg += a_name;
    msg += "'";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  return NULL;
}

Namespace* Node::namespace_find(const std::string& name, const bool exceptionWhenNotFound)
throw(RuntimeException) {
  Namespace* result = a_root->a_namespaces->find(name);

  if(result == NULL && exceptionWhenNotFound == true) {
    string msg(asString());
    msg += " | Namespace: ";
    msg += name;
    msg += " | Namespace is not defined";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  return  result;
}

string Node::asString() const
throw() {
  string result("xml::Node { Name: ");
  result += a_name;

  if(a_namespace != NULL) {
    result += " | ";
    result += a_namespace->asString();
  }

  return result += " }";
}

const Attribute* Node::find(const char* attrName, const_attribute_iterator begin, const_attribute_iterator end)
throw() {
  Attribute* attribute;

  for(; begin != end; begin ++) {
    attribute = *begin;

    if(anna_strcmp(attribute->getName(), attrName) == 0)
      return attribute;
  }

  return NULL;
}

